version = '0.9.5'

# As of v0.9.4 there should be a separate
# dataversion. in case the dataset structure
# has not changed, it is an overkill and
# to require to reload previously cached datasets.
#
# The dataversion has to be bumped whenever,
# the cached files are not compatible anymore with
# the new version anymore.
dataversion = '0.9.4'
